"""Utility class for exporting BigQuery data."""

from time import time
from datetime import timedelta
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from functools import cache

import polars as pl

from google.cloud.storage import Client as StorageClient, Bucket, Blob
from google.cloud.bigquery import Client as BQClient, Table, ExtractJobConfig

from neoval_py_utils.utils import (
    get_env_or_default,
    gcs_blobs_to_df,
    prepare_query_hash,
)
from typing import Optional


class Exporter:
    """A wrapper around a BQ query facilitating fast export and local caching.

    For efficient BigQuery data export and caching to prevent redundant query
    executions.

    The class can be used as a singleton: it should be instantiated
    once, and the same instance can repeatedly have queries passed to it.

    For simplicity, export method is aliased by the `<` operator.

    ```py
    exporter = Exporter()

    # `df` would contain a cached copy of query results - subsequent calls
    # with the same query string would not actually execute the query.
    df = exporter < "SELECT * FROM `my-dataset.my-table`"

    # The same instance should be used for all queries.
    df = exporter < "SELECT x, y FROM `my-dataset.my-other-table`"
    ```

    Limitations (Will be addressed in future versions):
    - The class is not thread-safe.
    - Does not check if the table it is exporting has been updated since cached.
    """

    def __init__(
        self,
        cache_max_age: Optional[float] = None,
        cache_bucket: Optional[str] = None,
        cache_dir: Optional[str] = None,
        bq_client: BQClient = None,
        gcs_client: StorageClient = None,
    ) -> None:
        """Initialise the Exporter class.

        Args:
            cache_max_age: Hours to cache local files.
            cache_bucket: Google Cloud Storage bucket name.
            cache_dir: Path to the local cache directory. Default is None.
            bq_client: Bigquery client.
            gcs_client: Google storage client.
        """
        # Default values
        self.cache_max_age = get_env_or_default(
            cache_max_age, 12, env_var="CACHE_MAX_AGE"
        )
        self.cache_dir = get_env_or_default(cache_dir, None, env_var="CACHE_DIR")
        self.cache_bucket = get_env_or_default(
            cache_bucket, "py_utils_cache", env_var="CACHE_BUCKET"
        )
        self.gcs_client: StorageClient = get_env_or_default(gcs_client, StorageClient())
        self.bq_client: BQClient = get_env_or_default(bq_client, BQClient())
        self.cache_bucket: Bucket = self.gcs_client.bucket(self.cache_bucket)

        # Setting up gcs buckets
        if not self.cache_bucket.exists():
            self.cache_bucket: Bucket = self.gcs_client.create_bucket(
                self.cache_bucket,
                location="US",
            )

            self.cache_bucket.add_lifecycle_delete_rule(age=1)
            self.cache_bucket.patch()

        # Setting up local cache directory
        self.cache_dir = None if self.cache_dir is None else Path(self.cache_dir)
        if self.cache_dir is not None:
            self.cache_dir.mkdir(parents=True, exist_ok=True)

    def export(self, query: str) -> pl.DataFrame:
        """Export a query to a parquet file, and return a polars DataFrame."""
        query_hash, table_ref = prepare_query_hash(query)

        cache_path = self._cache_path(query_hash)

        if cache_path is not None:
            self._cleanup_cache()
            if cache_path.exists():
                delta = time() - cache_path.stat().st_mtime
                if timedelta(hours=int(self.cache_max_age)) > timedelta(seconds=delta):
                    df = pl.read_parquet(cache_path)
                    return df

        if table_ref is None:
            job = self.bq_client.query(query)
            table_ref = job.result()._table

        blobs = self.bq_to_gcs(table_ref, query_hash, deterministic=False)

        df = gcs_blobs_to_df(blobs)

        if cache_path is not None:
            df.write_parquet(cache_path)

        return df

    def __lt__(self, query: str) -> pl.DataFrame:
        """Set the `less than` operator to the export method."""
        return self.export(query)

    @cache
    def _cache_path(self, name: str) -> Path | None:
        """Get the path to the cached parquet file."""
        if self.cache_dir is None:
            return None

        return self.cache_dir / f"{name}.parquet"

    def _cleanup_cache(self) -> None:
        fs = self.cache_dir.glob("*.parquet")
        for f in fs:
            delta = time() - f.stat().st_mtime
            if timedelta(seconds=delta) > timedelta(hours=int(self.cache_max_age)):
                f.unlink()

    def _list_blobs(self, name: str) -> list[Blob]:
        """List all blobs in the cache bucket with the given prefix."""
        i = self.gcs_client.list_blobs(self.cache_bucket, prefix=f"{name}")
        return list(i)

    def _delete_blobs_in_bucket(self, name: str) -> None:
        """Delete all blobs in the cache bucket with the given prefix."""
        with ThreadPoolExecutor(24) as executor:
            executor.map(
                lambda blob: blob.delete(),
                self._list_blobs(name),
            )

    def bq_to_gcs(
        self, table_ref: str, folder_name: str | None = None, deterministic: bool = True
    ) -> list[Blob]:
        """Export a BigQuery table to a GCS bucket.

        Args:
            table_ref: Bigquery table reference
            folder_name: Name of folder in GCS bucket
            deterministic: If true, data is ordered by a hash of all columns

        Returns: List of blobs in the GCS bucket that were result of the export

        """
        if folder_name is None:
            folder_name = table_ref.split(".")[-1]

        self._delete_blobs_in_bucket(folder_name)

        table: Table = self.bq_client.get_table(table_ref)
        location: str = (
            f"gs://{self.cache_bucket.name}/{folder_name}/{folder_name.split('/')[-1]}_*"
        )

        if deterministic:
            query = f"""
                EXPORT DATA OPTIONS (
                    uri='{location}',
                    format='PARQUET',
                    overwrite=true) AS
                SELECT *
                FROM {table} AS t
                ORDER BY FARM_FINGERPRINT(TO_JSON_STRING(t))
            """
            self.bq_client.query(query).result()

        else:
            self.bq_client.extract_table(
                table,
                [location],
                job_config=ExtractJobConfig(
                    destination_format="PARQUET",
                ),
            ).result()

        return self._list_blobs(folder_name)



"""Events for the build process."""

from dataclasses import dataclass
from enum import Enum


class IPDBBuildEventType(Enum):
    """Enum of the event types that can occur during the build process."""

    SOURCE_DATASET_METADATA_RETRIEVED = 1
    PULLING_SOURCE_TABLE = 2
    DUCKDB_TABLE_CREATED = 3
    SQLITE_TABLE_CREATED = 4
    BUILD_COMPLETE = 5
    UPLOAD_COMPLETE = 6
    PREPARE_SCRIPTS_COMPLETE = 7
    FAILURE = 8


@dataclass
class IPDBBuildEvent:
    """Event class for the build process."""

    type: IPDBBuildEventType
    data: dict



"""In process (embedded) Database Builder class."""

import concurrent.futures
import datetime

from typing import Optional, ClassVar

import json
from json import JSONDecodeError
from textwrap import dedent
import os
from pathlib import Path

import sqlite3
from sqlite3 import Connection as SQLiteConnection
import yaml

from google.cloud.bigquery import Client as BigQueryClient
from google.cloud.bigquery.table import Table as ArrowTable
from google.cloud.storage import Client as StorageClient
from google.cloud.storage import Blob

from pyarrow import RecordBatch

import duckdb
from duckdb import DuckDBPyConnection

from neoval_py_utils.exporter import Exporter
from neoval_py_utils.ipdb_build_event import IPDBBuildEvent, IPDBBuildEventType
from neoval_py_utils.template import Template
from neoval_py_utils.utils import gcs_blob_to_table

_EventTypes = IPDBBuildEventType
_Event = IPDBBuildEvent


class IPDBBuilder:
    """In process Database Builder class.

    A class for constructing local SQLite3 and DuckDB databases from the contents
    of a Google Bigquery dataset.
    """

    def __init__(self, dataset: str, project: str) -> None:  # noqa: D107
        self.dataset = dataset
        self.project = project
        self._client: BigQueryClient = None

        self._source_dataset_metadata: list[dict] = None

        self._exporter = Exporter()

    BQ_SQLITE_TYPE_MAP: ClassVar = {
        "INT64": "INTEGER",
        "FLOAT64": "REAL",
        "STRING": "TEXT",
        # These types don't exist in SQlite - but should logically be serialized
        # as strings. They don't need to be converted on the way back out - since
        # JSON itself cannot encode these types.
        "DATE": "TEXT",
        "DATETIME": "TEXT",
        "TIMESTAMP": "TEXT",
        "BOOL": "BOOLEAN",  # The SQLite type affinity is `NUMERIC`.
        # This doesn't exist in SQlite, but will be serialized as a WKT string.
        # It cannot / should not be compared / sorted / transformed in a query.
        #
        # NOTE: This _could_ be done with an adapter, but that would not be
        # a solution for DuckDB... the most consistent solution to the
        # serialization of geography types is to always store (in the BQ source
        # dataset) x / y pairs for `POINT()` types and GeoJSON encoding for
        # all shapes / lines.
        "GEOGRAPHY": "GEOGRAPHY",
        # NOTE: These types don't natively exist in SQLite, but we can add a
        # dynamic type that will allow us to serialize them as integers or
        # strings automagically.
        #
        # SEE: `https://docs.python.org/3/library/sqlite3.html#sqlite3.register_adapter`.
        #
        # Similarly, any API response code should register converters so that these
        # values are converted to the logical Python type prior to
        # JSON serialization of the response data.
        "STRUCT": "COMPOSITE",
        "ARRAY": "COMPOSITE",
    }

    @staticmethod
    def _composite_adapter(value: dict | list | tuple) -> str:
        """Composite adapter.

        Converts any value that is landing in a `COMPOSITE` type column in
        SQLite into a string (JSON) prior to insertion (see docs in the type
        mapping above).
        """
        return json.dumps(value, sort_keys=False)

    @property
    def client(self) -> BigQueryClient:
        if not self._client:
            self._client = BigQueryClient()

        return self._client

    @staticmethod
    def _parse_table_metadata(table_metadata: dict) -> dict:
        # Table descriptions are sometimes (for some reason) JSON encoded -
        # this works around that, while also handling the value `None`.
        try:
            description = table_metadata["description"]
            if description is not None:
                table_metadata["description"] = json.loads(
                    table_metadata["description"]
                )
        except JSONDecodeError:
            pass

        # Remove the `description` key where it has no value.
        columns = table_metadata.get("columns")
        for column in columns:
            if column.get("description") is None:
                column.pop("description")

        return table_metadata

    @property
    def source_dataset_metadata(self) -> list:
        if self._source_dataset_metadata:
            return self._source_dataset_metadata

        query = dedent(
            """\
                WITH table_name AS (
                    SELECT DISTINCT
                        table_name AS name
                    FROM {project}.{dataset}.INFORMATION_SCHEMA.TABLES
                    WHERE table_type = 'BASE TABLE'
                ),

                table_description AS (
                    SELECT
                        table_name AS name,
                        option_value AS description
                    FROM
                        {project}.{dataset}.INFORMATION_SCHEMA.TABLE_OPTIONS table_option
                    WHERE UPPER(option_name) = 'DESCRIPTION'
                ),

                table AS (
                    SELECT name, NULLIF(description, '""') AS description,
                    FROM table_name LEFT JOIN table_description USING (name)
                ),

                table_column AS (
                    SELECT
                        tc.table_name AS name,
                        ARRAY_AGG(
                            STRUCT(
                                tc.column_name AS name,
                                tc.data_type AS type,
                                tfp.description AS description
                            ) ORDER BY tc.ordinal_position
                        ) AS columns
                    FROM
                        {project}.{dataset}.INFORMATION_SCHEMA.COLUMNS tc LEFT JOIN
                        {project}.{dataset}.INFORMATION_SCHEMA.COLUMN_FIELD_PATHS tfp USING (
                            table_name,
                            column_name
                        )
                    GROUP BY tc.table_name
                )

                SELECT * FROM table LEFT JOIN table_column USING (name) ORDER BY name
            """
        ).format(project=self.project, dataset=self.dataset)

        self._source_dataset_metadata = [
            self._parse_table_metadata(dict(r))
            for r in self.client.query(query).result()
        ]

        return self._source_dataset_metadata

    @staticmethod
    def _normalize_config(val: list | str) -> list:
        if val is None:
            return []

        if isinstance(val, str):
            return [val]

        return val

    @staticmethod
    def _merge_config_and_dataset_metadata(
        dataset_metadata: list[dict], config: list[dict]
    ) -> list[list[dict]]:
        """Takes the provided config and adds metadata from BigQuery.

        Returns a list of table metadata for both duckdb and sqlite.
        """
        if "duckdb" in config:
            _config = {
                table_config["name"]: {
                    "primary_key": IPDBBuilder._normalize_config(
                        table_config.get("primary_key", [])
                    )
                }
                for table_config in (config["duckdb"] or [])
            }

            _dataset_metadata = json.loads(json.dumps(dataset_metadata))

            # NOTE: This mutates the above variable (hence the copy).
            for table_metadata in _dataset_metadata:
                name: str = table_metadata["name"]
                table_config: dict = _config.get(
                    name,
                    {
                        "primary_key": [],
                    },
                )

                table_metadata.update(table_config)

            # This should remove the tables that are not in the config
            duckdb_metadata = [
                table_metadata
                for table_metadata in _dataset_metadata
                if table_metadata["name"] in _config
            ]
        else:
            duckdb_metadata = None

        if "sqlite" in config:
            _config = {
                table_config["name"]: {
                    "primary_key": IPDBBuilder._normalize_config(
                        table_config.get("primary_key", [])
                    )
                }
                for table_config in (config["sqlite"] or [])
            }

            _dataset_metadata = json.loads(json.dumps(dataset_metadata))

            # NOTE: This mutates the above variable (hence the copy).
            for table_metadata in _dataset_metadata:
                name: str = table_metadata["name"]
                table_config: dict = _config.get(
                    name,
                    {
                        "primary_key": [],
                    },
                )

                table_metadata.update(table_config)

            # This should remove the tables that are not in the config
            sqlite_metadata = [
                table_metadata
                for table_metadata in _dataset_metadata
                if table_metadata["name"] in _config
            ]
        else:
            sqlite_metadata = None

        return [duckdb_metadata, sqlite_metadata]

    @staticmethod
    def _render_duckdb_create_table_statement(
        con: DuckDBPyConnection,
        target_name: str,
        source_name: str = "temp",
        primary_key: list[str] = [],
    ) -> str:
        declarations = (
            con.execute(
                dedent(
                    """\
                        SELECT
                            CONCAT(column_name, ' ', data_type) AS declaration,
                        FROM INFORMATION_SCHEMA.COLUMNS
                        WHERE table_name = '{}'
                        ORDER BY ordinal_position ASC
                    """
                ).format(source_name)
            )
            .arrow()
            .to_pylist()
        )

        pk = (
            ""
            if not primary_key
            else ",\n\n    PRIMARY KEY ({})".format(", ".join(primary_key))
        )

        return (
            f"CREATE OR REPLACE TABLE {target_name} (\n"
            + (",\n".join(["    " + v["declaration"] for v in declarations]))
            + pk
            + "\n)"
        )

    @staticmethod
    def _render_sqlite_column_declaration(column: dict) -> str:
        name = column["name"]
        dtype = IPDBBuilder.BQ_SQLITE_TYPE_MAP.get(column["type"], "UNKNOWN")

        return f"{name} {dtype}"

    @staticmethod
    def _render_sqlite_create_table_statement(table_metadata: list[dict]) -> str:
        name = table_metadata["name"]
        columns = table_metadata["columns"]

        primary_key = table_metadata.get("primary_key", [])
        pk = (
            ""
            if not primary_key
            else ",\n\n    PRIMARY KEY ({})".format(", ".join(primary_key))
        )

        return (
            f"CREATE TABLE {name} (\n"
            + (
                ",\n".join(
                    [
                        "    " + IPDBBuilder._render_sqlite_column_declaration(column)
                        for column in columns
                    ]
                )
            )
            + pk
            + "\n)"
        )

    @staticmethod
    def _prepare_directory_and_pathname(path_ref: str | Path, pathname: str) -> Path:
        """Prepare pathname for build method.

        Ensure that `path_ref` exists, and that the specified pathname within it
        does not. If those files / folders exist, they are deleted - so that the
        returned `Path`s always reference files that don't exist on disk.
        """
        p: Path = Path(path_ref)

        if not p.exists():
            p.mkdir()

        pathname = p / pathname
        if pathname.exists():
            pathname.unlink()

        return pathname

    @staticmethod
    def _verify_directory_and_pathname(path_ref: str | Path, pathname: str) -> Path:
        """Prepare pathnames for prepare method.

        Ensure that `path_ref` and the pathnames exist, otherwise raises an error.
        Opposite to the method above, this is used in the `prepare` method to ensure
        that the databases we want to operate on have already been created.
        """
        p: Path = Path(path_ref)

        if not p.exists():
            raise ValueError(f"{path_ref} directory does not exist.")

        pathname = p / pathname
        if not pathname.exists():
            raise FileNotFoundError(f"{pathname} does not exist.")

        return pathname

    def _insert_data(
        self, blob: Blob, cur: DuckDBPyConnection, name: str, i: int
    ) -> None:
        attempt = 1
        while attempt < 6:  # noqa: PLR2004
            try:
                t: ArrowTable = gcs_blob_to_table(blob)  # noqa: F841
                cur.execute(f"INSERT INTO {name} SELECT * FROM t")
                return "SUCCESS"
            except:  # noqa: E722
                attempt += 1
        return f"ERROR: blob {i}"

    def _create_duckdb_table(
        self, table_metadata: dict, duckdb_con: DuckDBPyConnection
    ) -> IPDBBuildEvent:
        cur = duckdb_con.cursor()

        name = table_metadata["name"]
        primary_key = table_metadata["primary_key"]

        blobs: list[Blob] = self._exporter.bq_to_gcs(
            f"{self.project}.{self.dataset}.{name}"
        )

        temp: ArrowTable = gcs_blob_to_table(blobs[0])  # noqa: F841
        cur.execute(
            f"CREATE OR REPLACE TABLE temp_{name} AS SELECT * FROM temp LIMIT 1000"
        )

        duck_db_create_table_statement = self._render_duckdb_create_table_statement(
            cur,
            target_name=name,
            source_name=f"temp_{name}",
            primary_key=primary_key,
        )

        cur.execute(duck_db_create_table_statement)
        # Download each parquet file from GCS at a time and insert into duckdb
        # 14/6/24 Parallel inserts would intermittently hang with duckdb versions >= 0.10.0.
        # At present setting max_worker to 1 increased build by 40s, please revisit in the future.
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            futures = [
                executor.submit(self._insert_data, blob, cur, name, i)
                for i, blob in enumerate(blobs)
            ]

            for future in concurrent.futures.as_completed(futures):
                if future.result().startswith("ERROR"):
                    return _Event(
                        _EventTypes.FAILURE,
                        data={
                            "name": name,
                        },
                    )

        cur.execute(f"DROP TABLE temp_{name}")

        return _Event(
            _EventTypes.DUCKDB_TABLE_CREATED,
            data={
                "name": name,
                "create_table_statement": duck_db_create_table_statement,
            },
        )

    def _create_sqlite_table(
        self, table_metadata: dict, sqlite_con: SQLiteConnection
    ) -> IPDBBuildEvent:
        table_data: ArrowTable = self._exporter.export(
            f"{self.project}.{self.dataset}.{table_metadata['name']}"
        ).to_arrow()
        name = table_metadata["name"]
        sqlite_create_table_statement = self._render_sqlite_create_table_statement(
            table_metadata
        )

        sqlite_con.execute("BEGIN TRANSACTION")
        sqlite_con.execute(sqlite_create_table_statement)

        b: RecordBatch
        for b in table_data.to_batches(500):
            rows = [tuple(r.values()) for r in b.to_pylist()]

            q = ", ".join(["?" for _ in rows[0]])
            sqlite_con.executemany(f"INSERT INTO {name} VALUES({q})", rows)

        sqlite_con.commit()

        return _Event(
            _EventTypes.SQLITE_TABLE_CREATED,
            data={
                "name": name,
                "create_table_statement": sqlite_create_table_statement,
            },
        )

    def _build_sqlite(self, sqlite_path: Path, metadata: list) -> None:
        sqlite_con: SQLiteConnection = sqlite3.connect(str(sqlite_path))
        try:
            sqlite3.register_adapter(dict, IPDBBuilder._composite_adapter)
            sqlite3.register_adapter(list, IPDBBuilder._composite_adapter)

            sqlite_con.execute("PRAGMA journal_mode = WAL;")
            sqlite_con.execute("PRAGMA cache_size = -5000;")
            sqlite_con.execute("PRAGMA synchronous = NORMAL")

            for table in metadata:
                yield self._create_sqlite_table(table, sqlite_con)

            sqlite_con.execute("VACUUM;")
        finally:
            sqlite_con.close()

    def _build_duckdb(self, duckdb_path: Path, metadata: list) -> None:
        duckdb_con: DuckDBPyConnection = duckdb.connect(str(duckdb_path))
        try:
            for table in metadata:
                yield self._create_duckdb_table(table, duckdb_con)

        finally:
            duckdb_con.close()

    def build(
        self,
        directory: str | Path,
        config: list[dict],
        upload_bucket: Optional[str] = None,
    ) -> None:
        (
            duckdb_metadata,
            sqlite_metadata,
        ) = self._merge_config_and_dataset_metadata(
            self.source_dataset_metadata, config
        )
        """
        Build the In-process duckdb and sqlite databases from a BQ dataset.

        Args:
            directory: Path to the directory where the database files will be stored
            config: Configuration for the database.
            upload_bucket: The GCS bucket to upload the db files to.
        Yields:
            _Event: IPDBBuildEvent indicating the progress or completion of the build process.
        """

        yield _Event(
            _EventTypes.SOURCE_DATASET_METADATA_RETRIEVED,
            data={"metadata": None},
        )

        pathnames = []

        if duckdb_metadata:
            duckdb_path = self._prepare_directory_and_pathname(
                directory, "main.duck.db"
            )
            pathnames.append(duckdb_path)

            events = self._build_duckdb(duckdb_path, duckdb_metadata)
            for event in events:
                yield event
            yield _Event(_EventTypes.BUILD_COMPLETE, data={"name": "duckdb"})

        if sqlite_metadata:
            sqlite_path = self._prepare_directory_and_pathname(
                directory, "main.sqlite.db"
            )
            pathnames.append(sqlite_path)

            events = self._build_sqlite(sqlite_path, sqlite_metadata)
            for event in events:
                yield event
            yield _Event(_EventTypes.BUILD_COMPLETE, data={"name": "sqlite"})

        if upload_bucket:
            bucket = StorageClient().bucket(upload_bucket)
            for path in pathnames:
                blob = bucket.blob(f"databases/{'/'.join(str(path).split('/')[-2:])}")

                if blob.exists():
                    bucket.copy_blob(
                        blob,
                        bucket,
                        f"databases/market/archive/{int(datetime.datetime.now().timestamp())}_{'/'.join(str(path).split('/')[-1:])}",
                    )

                blob.upload_from_filename(str(path), timeout=300)
                yield _Event(
                    _EventTypes.UPLOAD_COMPLETE,
                    data={"name": str(path).rsplit("/", maxsplit=1)[-1]},
                )

    def prepare(self, db_directory: str | Path, template_path: str | Path) -> None:
        """Run prepare scripts.

        This method assumes there is a `duckdb` and/or `sqlite` folder inside the
        `template_path`, which is used to concat multiple queries and may contain
        macros. The individual queries that need to be run to create views should
        also exist within a `prepare` folder. For example:

        app/templates/duckdb/prepare.sql
        app/templates/duckdb/prepare/view1.sql
        app/templates/duckdb/prepare/view2.sql
        """
        env = Template(template_path)

        if os.path.exists(f"{template_path}/duckdb/prepare.sql"):
            duckdb_path = self._verify_directory_and_pathname(
                db_directory, "main.duck.db"
            )
            duckdb_con: DuckDBPyConnection = duckdb.connect(str(duckdb_path))
            files = os.listdir(f"{template_path}/duckdb/prepare")

            duckdb_query = env.get("prepare.sql", "duckdb").render(files=files)

            try:
                duckdb_con.execute(duckdb_query)

                r = duckdb_con.execute(
                    "SELECT table_name, table_type FROM information_schema.tables"
                )

                yield _Event(
                    _EventTypes.PREPARE_SCRIPTS_COMPLETE,
                    data={
                        "database": "duckdb",
                        "tables": yaml.safe_dump(r.arrow().to_pylist(), indent=4),
                    },
                )

            finally:
                duckdb_con.close()

        if os.path.exists(f"{template_path}/sqlite/prepare.sql"):
            sqlite_path = self._verify_directory_and_pathname(
                db_directory, "main.sqlite.db"
            )
            sqlite_con: SQLiteConnection = sqlite3.connect(str(sqlite_path))
            try:
                files = os.listdir(f"{template_path}/sqlite/prepare")
                sqlite_query = env.get("prepare.sql", "sqlite").render(files=files)

                sqlite_con.executescript(sqlite_query)
                sqlite_con.commit()
                r = sqlite_con.execute(
                    "SELECT type, tbl_name FROM sqlite_master WHERE type IN ('table', 'view')"
                )
                rows = r.fetchall()

                data = [{"table_name": row[1], "table_type": row[0]} for row in rows]

                yield _Event(
                    _EventTypes.PREPARE_SCRIPTS_COMPLETE,
                    data={
                        "database": "sqlite",
                        "tables": yaml.safe_dump(data, indent=4),
                    },
                )
            finally:
                sqlite_con.close()



"""Utility functions for the neoval-py-utils package."""

from os import environ
from textwrap import dedent
from typing import Any, Optional

from google.cloud.storage import Blob
from pyarrow import Table as PyArrowTable
import polars as pl
from io import BytesIO
from hashlib import sha1
from functools import cache
from concurrent.futures import ThreadPoolExecutor
import yaml


@cache
def bq_hash(query: str) -> str:
    """Generate a hash from the query that is requested from bq and used for the local cache."""
    h = sha1()
    h.update(query.encode())
    return h.hexdigest()[:12]


def prepare_query_hash(query: str) -> tuple[str, str | None]:
    """Prepare the query and return the hash and the table reference.

    Args:
        query: Query or table name

    Returns:
        Tuple of the query hash and the table reference, if a query is given, None is returned.

    """
    query = dedent(query).strip()

    is_select_all_table = not any(char.isspace() for char in query)
    table_ref = None

    if is_select_all_table:
        table_ref = query
        query = f"SELECT * FROM `{query}`"

    return bq_hash(query), table_ref


def get_env_or_default(
    attribute: Any, default: Any, env_var: Optional[str] = None
) -> Any:
    """Return the attribute if not None, environment variable or default.

    Args:
        attribute: Method variable input.
        default: The default value if there is no environment variable.
        env_var: The name of the environment variable to check. Optional.

    Returns:
        The value of the attribute(function input) > environment variables > default value.
    """
    if attribute is not None:
        return attribute
    if env_var is not None:
        return environ.get(env_var.upper(), default)
    return default


def gcs_blobs_to_df(blobs: list[Blob]) -> pl.DataFrame:
    """Download all blobs in the cache bucket with the given prefix/table name.

    Args:
        blobs: List of blobs to download.

    Returns: List of bytes for each blob.

    """
    with ThreadPoolExecutor(24) as executor:
        blob_bytes = executor.map(
            lambda blob: BytesIO(blob.download_as_bytes()),
            blobs,
        )

    return pl.concat([pl.read_parquet(s) for s in blob_bytes])


def gcs_blob_to_table(blob: Blob) -> PyArrowTable:
    """Download a blob from GCS and convert to a PyArrowTable."""
    b = BytesIO(blob.download_as_bytes())
    df = pl.read_parquet(b)
    return df.to_arrow()


def fdump(data: list[dict]) -> str:
    """Format and dump the data to a yaml string."""
    yaml.SafeDumper.og_represent_str = yaml.SafeDumper.represent_str

    def str_representer(dumper, data):  # noqa: ANN001, ANN202
        if "\n" in data or len(data) > 20:  # noqa: PLR2004
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=">")
        return dumper.og_represent_str(data)

    yaml.add_representer(str, str_representer, Dumper=yaml.SafeDumper)

    return yaml.safe_dump(data=data, sort_keys=False, indent=4, width=80)



"""A command line tool to build an in process DuckDB and/or SQLite database from BQ datasets."""

from sys import stderr
from pathlib import Path
import traceback

import yaml
import typer

from neoval_py_utils.ipdb_builder import (
    IPDBBuilder,
    IPDBBuildEventType,
    IPDBBuildEvent,
)
from neoval_py_utils.utils import fdump

app = typer.Typer()


@app.command()
def make_config(dataset: str, project: str) -> str:
    """Prints a default configuration to be used with the build command.

    The final config file should include the list of tables for each database.
    For example:

    sqlite:
        -   name: canonical_address
            primary_key: null
        -   name: sa2_locality_correspondence_2021
            primary_key: null
        -   name: gnaf_core
            primary_key: null
    duckdb:
        -   name: gnaf_core
            primary_key: null
    """
    b = IPDBBuilder(dataset=dataset, project=project)

    s = fdump(
        [
            {"name": table_metadata["name"], "primary_key": None}
            for table_metadata in b.source_dataset_metadata
        ]
    )

    print(s)
    return s


def _filter_table_metadata(table_metadata: dict) -> dict:
    name: str = table_metadata["name"]
    description: str = table_metadata.get("description", None)

    table_metadata = {"name": name}
    if description:
        table_metadata["description"] = description

    return table_metadata


@app.command()
def build(
    dataset: str,
    project: str,
    directory: Path,
    config: Path,
    upload_bucket: str = typer.Option(None),
) -> None:
    """Build the in process database(s)."""
    b = IPDBBuilder(dataset=dataset, project=project)
    directory = directory.resolve()
    directory.mkdir(parents=True, exist_ok=True)

    try:
        config = config.resolve()
        with config.open("r") as file:
            config = yaml.safe_load(file)

        # List all the tables built in either database
        all_tables = [item["name"] for sublist in config.values() for item in sublist]

        events = b.build(
            directory=directory.resolve(),
            config=config,
            upload_bucket=upload_bucket,
        )

        event: IPDBBuildEvent
        code = 0
        for event in events:
            if event.type == IPDBBuildEventType.FAILURE:
                code = 1
            print(event.type, event.data.get("name"), file=stderr)
            create_table_statement = event.data.get("create_table_statement", None)

            if create_table_statement:
                print(create_table_statement, "\n", file=stderr)
                continue

        with (directory / "main.metadata.yaml").open("w") as f:
            f.write(
                fdump(
                    [
                        _filter_table_metadata(table_metadata)
                        for table_metadata in b.source_dataset_metadata
                        if table_metadata["name"] in all_tables
                    ]
                )
            )

        if code == 1:
            raise typer.Exit(code=code)

    except Exception as e:
        traceback.print_exc()
        raise e


@app.command()
def prepare(
    dataset: str, project: str, db_directory: Path, template_path: Path
) -> None:
    """Run scripts to add views/virtual tables/etc. to the database(s)."""
    b = IPDBBuilder(dataset=dataset, project=project)

    try:
        events = b.prepare(
            db_directory=db_directory.resolve(), template_path=template_path.resolve()
        )

        event: IPDBBuildEvent
        for event in events:
            print(event.type, event.data.get("database"), file=stderr)
            print(event.data.get("tables"), "\n", file=stderr)

    except Exception as e:
        traceback.print_exc()
        typer.Exit(code=1)
        raise e


if __name__ == "__main__":
    app()



"""Template class for rendering SQL for existing In Process databases."""

from typing import Iterable, Union, Optional
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, select_autoescape, TemplateNotFound
from jinja2.environment import Template


class Template:
    """A singleton class for rendering SQL templates.

    Example usage:
        template_path = Path.cwd() / "templates"
        template = Template(template_path)
        query = template.get("example.sql").render()
    """

    def __new__(cls, *args) -> None:  # noqa: ANN002
        if not hasattr(cls, "instance"):
            cls.instance = super(Template, cls).__new__(cls)  # noqa: UP008
        return cls.instance

    def __init__(self, templates_path: str | Path) -> None:
        """Initialize the class with a jinja environment."""
        if not hasattr(self, "env"):
            templates_path = Path(templates_path)

            self.env = Environment(
                loader=FileSystemLoader(templates_path),
                autoescape=select_autoescape(
                    disabled_extensions=("sql"),
                    default_for_string=False,
                    default=False,
                ),
                trim_blocks=True,
                lstrip_blocks=True,
                enable_async=True,
            )

            self.env.tests["populated"] = self._populated
            self.env.filters["listify"] = self._listify

    def get(self, name: str, db: Optional[str] = None) -> Template:
        """A method to retrieve the `Template`.

        If no database is passed, it will look for the named template in both
        subfolders. Returns a jinja `Template` object that can then be rendered.
        """
        if not db:
            try:
                for template in [f"duckdb/{name}", f"sqlite/{name}", name]:
                    return self.env.get_template(template)
            except TemplateNotFound as exc:
                raise ValueError(f"`{name}` does not exist.") from exc

        elif db not in ("duckdb", "sqlite"):
            raise ValueError(f"`{db}` is not a valid database value.")

        else:
            try:
                return self.env.get_template(f"{db}/{name}")
            except TemplateNotFound:
                try:
                    return self.env.get_template(name)
                except TemplateNotFound as exc:
                    raise ValueError(
                        f"`{name}` does not exist in `{db}` templates."
                    ) from exc

    @staticmethod
    def _populated(value: Union[Iterable, float, int, str]) -> bool:
        """A custom test.

        This is a custom test - it's designed to be used in conjunction with
        `listify` (or the like), which is below.

        ```sql
        {% if values is populated %}
        ```

        Will evaluate to `True` where:
        - `values` is a primitive value, or
        - `values` is an iterable with at least one item in it.
        """
        if isinstance(value, (float, int, str)):
            return True

        if isinstance(value, Iterable):
            return len(value) > 0

        return False

    @staticmethod
    def _listify(value: Union[Iterable, float, int, str]) -> str:
        """A custom filter to generate lists of strings.

        This is a custom filter used to generate lists of strings for use in, for
        instance, an `IN (...)` clause. It can be called from Jinja2 using something
        like:

        ```sql
        # values = ["a", "b", "c"]
        AND value IN ({{ values|listify }})

        # Results in `AND value IN ("a", "b", "c")`
        ```

        Importantly, this filter should only be used where the inputs are intended
        to be serialized as strings (and so surrounded by quotes) - it's equivalent
        to (but less verbose that) calling something like
        `{{ values|map('tojson')|join(', ') }} }}`.

        If you need an equivalent list that is not quoted, you can just called
        `{{ values|join(', ') }}`
        """
        if not isinstance(value, (str, Iterable)):
            value = [value]

        return ", ".join([f'"{v}"' for v in value])
